#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from smtplib          import SMTP
from email.message    import EmailMessage
from cve_manager.conf import MailConfParser, SMTP_CON_SEC, RECIPIENTS_SEC

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Mailer():

	def __init__(self, conf_file_path):

		self.conn = {}
		self.recipients = {}
		self.critical_err = ''

		# Reading the conf
		conf, critical_err = MailConfParser(conf_file_path).Parse()
		if critical_err:
			return

		# Getting values of the parameters
		self.conn, self.recipients = conf.Get([SMTP_CON_SEC, RECIPIENTS_SEC])
		for p in ('host', 'port'):
			if self.conn.get(p):
				continue
			self.critical_err += ('; ' if self.critical_err else '') + \
				f'Can\'t find "{p}" in SMTP connection params'
		if not self.recipients:
			self.critical_err += ('; ' if self.critical_err else '') + \
				'There are no email recipients'

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send an email using params from the conf-file

	def Execute(self, category_of_recipients, email_text, type_of_issues=None,
			branch=None, distro=None):

		if self.critical_err:
			return self.critical_err

		if category_of_recipients not in self.recipients:
			err = 'Unknown category of email recipients ' + \
				category_of_recipients
			return err

		# Forming a subject of this email
		subj = f'[cve-monitor] {category_of_recipients}'
		if type_of_issues:
			subj += f':{type_of_issues}'
		if branch:
			subj += f':{branch}'
			if distro:
				subj += f':{distro}'

		# Creating a text/plain message
		msg = EmailMessage()
		msg.set_content(email_text)
		msg['Subject'] = subj
		msg['From'] = self.conn.get('raddress', '')
		msg['To'] = self.recipients[category_of_recipients]

		# Send the message via our own SMTP server
		try:
			host = self.conn['host']
			port = self.conn['port']
			with SMTP(host, port) as s:
				s.starttls()
				usr = self.conn.get('user', '')
				passwd = self.conn.get('password', '')
				s.login(usr, passwd)
				s.send_message(msg)
		except Exception as exc:
			err = f'{msg["Subject"]}, {exc}'
			return err

		return ''
