#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from collections         import defaultdict
from mysql.connector     import Error as MySQLError
from ax.printer          import Printer
from cve_manager.defines import VUL_MARKERS
from cve_manager.common  import SplitVulId, ReviseCVEVolumes, DISPUTED_REGEX
from cve_manager.db      import DB, QueryToFetchIssues, PrefixOfHistoryTables
from cve_monitor.common  import PackagesList

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class FatalError(Exception):
	def __init__(self, msg=''):
		super(FatalError, self).__init__(msg)

def NoNVDNameErrMsg(branch, issues_table):
	return (f'There is an entry in the {issues_table} table for a package that '
		'is not mapped to any product name, you have to remap package names of '
		f'the {branch} branch, or recreate the {issues_table} table, or do both')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Condition():

	def __init__(self, txt='', type_of_issues=None, fix_status=None,
			packages=None, years=None, target=None):

		self.__condition = txt

		# Add specification prefix to a column name
		def Specify(column_name):
			return f'{target}.{column_name}' if target else column_name
		# - nested func

		if fix_status != None:
			column = Specify('fix')
			if fix_status == 'fixed':
				self.Add(f"{column} REGEXP '[ceb]'")
			elif fix_status == 'unfixed':
				self.Add(f"{column} = 'V'")
			elif fix_status == 'unclear':
				self.Add(f"{column} in ('L', '?')")

		if packages:
			column = Specify('name')
			packages = [f"'{p}'" for p in packages]
			self.Add(f"{column} in ({','.join(packages)})")

		if years:
			column = Specify('vul_year')
			self.Add(f"{column} in ({','.join(years)})")

	# Add given condition with a use of the 'AND' keyword
	# (the cnd is expected to be of type Condition or str)
	def Add(self, cnd):
		txt = cnd.What() if type(cnd) is Condition else cnd
		if not txt:
			return
		if not self.__condition:
			self.__condition = txt
		else:
			self.__condition += f' AND {txt} '

	# Form a WHERE part of a query using currrent condition
	def Where(self):
		return f'WHERE {self.__condition} ' if self.__condition else ''

	# Getter
	def What(self):
		return self.__condition

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Query():

	def __init__(self, mysql_config, printer=None):

		self.context = None
		self.printer = printer
		if not self.printer:
			self.printer = Printer(silent = True)

		# Connecting to the CVE database
		self.db = DB(self.printer)
		if not self.db.Connect(mysql_config):
			exit(1)

	def SetContext(self, context):

		if any(not hasattr(context, attr) for attr in ('branches', 'packages',
				'fix', 'years', 'vul_ids', 'order', 'patch', 'group', 'cure')):
			return False

		self.context = context

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a list of packages of a specified branch

	def __Packages(self, branch=None):

		return PackagesList(self.context.packages, branch)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Form a sorting part of a query

	def __CustomOrder(self, type_of_issues=None, branch=None, tables={},
			columns={}):

		order = ''

		if not tables and type_of_issues and branch:
			tables = {
				'packages': f'{branch}_src',
				'cvss_score_v2': f'{type_of_issues}_vul_import',
				'cvss_score_v3': f'{type_of_issues}_vul_import',
				'fix': f'{branch}_{type_of_issues}_issues'
				}

		if self.context.order:
			table = tables.get(self.context.order)
			order += 'ORDER BY '
			if self.context.order in columns:
				order += columns[self.context.order] + ' '
			else:
				order += f'{f"{table}." if table else ""}{self.context.order} '
			if self.context.order in ('cvss_score_v2', 'cvss_score_v3', 'fix'):
				order += 'DESC '

		return order

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check an existence of a given table in the DB

	def TableExists(self, table):
		return self.db.TableExists(table)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query all\unfixed\fixed issues for a given branch and (optionaly)
	#	package names and years

	def Issues(self, type_of_issues, seniors_of_the_branches):

		issues = defaultdict(list)
		suggestions = defaultdict(lambda: defaultdict(int))

		if not self.context:
			raise FatalError('Context is not defined')
			return issues, suggestions

		vul_import_table = f'{type_of_issues}_vul_import'
		vul_marker = VUL_MARKERS.get(type_of_issues, '')

		for branch in self.context.branches:

			issues_table = f'{branch}_{type_of_issues}_issues'
			src_table = f'{branch}_src'

			if any(not self.db.TableExists(t) for t in (issues_table, src_table)):
				continue

			# Forming a query to get particular issues

			package_basic_desc = [f'b.{el}' for el in
				('name', f'{type_of_issues}_name', 'ver', 'rel')]
			package_basic_desc_fields = ', '.join(package_basic_desc)
			scores = [f'c.cvss_score_{v}' for v in ('v2', 'v3')]
			scores_fields = ', '.join([f'MAX({s}) max_{s[2:]}' for s in scores]
				if self.context.group else scores)
			disputed = fr"REGEXP_LIKE(c.summary, {DISPUTED_REGEX})"
			patch = f', c.patch ' if self.context.patch else ''

			query  = "SELECT "
			query += "COUNT(a.vul_year), " if self.context.group else \
				f"CONCAT('{vul_marker}', a.vul_year, '-', a.vul_num) vul_id, "
			query += (f"{package_basic_desc_fields}, "
				"GROUP_CONCAT(DISTINCT a.vul_ver "
					"ORDER BY a.vul_ver ASC SEPARATOR ' '), "
				f"{scores_fields}, "
				f"{f'MAX({disputed})' if self.context.group else disputed}, "
				f"a.fix, b.acl {patch} FROM {issues_table} a "
				f"INNER JOIN {src_table} b "
					"ON a.package_id = b.id "
				f"INNER JOIN {vul_import_table} c "
					"ON (a.vul_year = c.vul_year AND a.vul_num = c.vul_num) ")

			if self.context.fix != "all" or self.context.packages or self.context.years:
				condition = Condition('', type_of_issues, self.context.fix,
					self.__Packages(branch), self.context.years)
				query += condition.Where()

			query += "GROUP BY "
			if self.context.group:
				query += f'{package_basic_desc_fields}, '
			else:
				query += 'a.vul_year, a.vul_num, ' \
					f'{package_basic_desc_fields}, ' \
					f'{scores_fields}, '
			order = self.__CustomOrder(
				tables={
					'cvss_score_v2': 'c',
					'cvss_score_v3': 'c',
					'fix': 'a',
					'package': 'b'
					},
				columns={
					col: f'max_{col}' for col in ('cvss_score_v2', 'cvss_score_v3')
					} if self.context.group else {}
				)
			query += f'a.fix, b.acl {patch} {order} '

			# Sending the query
			cursor = self.db.Cursor(query)

			if cursor == None:
				err = f'Can\'t get {branch} {type_of_issues.upper()} issues'
				raise FatalError(err)
				break

			if not cursor or cursor.rowcount < 1:
				continue

			for entry in cursor:
				if entry[2] == None:
					raise FatalError(NoNVDNameErrMsg(branch, issues_table))
					return issues, suggestions
				issues[branch].append(entry)

			# Getting fix-suggestions
			if self.context.cure and self.context.fix == 'unfixed':
				# Getting the diff between branches
				seniors_of_the_branch = seniors_of_the_branches.get(branch, [])
				compared_branches = ([branch] + [b for b in self.context.all_branches
					if b != branch and b in seniors_of_the_branch])
				if len(compared_branches) < 2:
					continue
				diff = self.CompareBranches(type_of_issues, compared_branches, False)
				# Forming a dict...
				#	...of number of fixed issues for a package in a branch
				if self.context.group:
					for compared_branch, diff_issues in diff.items():
						for diff_issue in diff_issues:
							suggestions[branch][(diff_issue[1], compared_branch)] += 1
				#	...of fix-flags for an issue in a branch
				else:
					for compared_branch, diff_issues in diff.items():
						for diff_issue in diff_issues:
							suggestions[branch][(diff_issue[0], compared_branch)] = 0

		return issues, suggestions

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query full desc of vulnerability with a given id

	def __QueryVulDesc(self, type_of_issues, vul_year, vul_num):

		table = type_of_issues + '_vul_import'
		vul_marker = VUL_MARKERS.get(type_of_issues, '')
		cols = [f"CONCAT('{vul_marker}', vul_year, '-', vul_num)",
			"software_list", "published", "modified", "cvss_score_v2",
			"cvss_score_v3", "cwe_id", "refs", "patch", "summary"]
		cond = f"vul_year = {vul_year} AND vul_num = '{vul_num}'"

		cursor = self.db.Select(cols, table, cond)

		if not cursor:
			return tuple()

		return cursor.fetchall()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query full desc and fix status for given vul ids

	def Entries(self, type_of_issues):

		import re

		entries = defaultdict(dict)

		if not self.context:
			raise FatalError('Context is not defined')
			return entries

		err = ''
		if not self.context.branches:
			err = 'Branches are not specified'
		elif not self.context.packages and not self.context.vul_ids:
			err = 'Packages and vul ids are not specified'
		if err:
			raise FatalError(err)
			return entries

		if not self.context.vul_ids:
			self.context.vul_ids = ['']

		# Filter excessive information from a given fix status
		def FilterFix(combined_fix):
			f = set(combined_fix)
			if 'V' in f:
				return 'V'
			if 'c' in f or 'e' in f:
				f -= f - {'c', 'e'}
			return ''.join(sorted(f))
		# - nested func

		for vul_id in self.context.vul_ids:
			for branch in self.context.branches:
				for package in \
						self.__Packages(branch) if self.context.packages else ['']:
					issues_table = f'{branch}_{type_of_issues}_issues'
					src_table = f'{branch}_src'
					if not self.db.TableExists(issues_table):
						continue
					query = (
						"SELECT a.vul_year, a.vul_num, b.name, "
							"GROUP_CONCAT(a.fix ORDER BY a.fix DESC SEPARATOR '') "
						f"FROM {issues_table} a "
							f"INNER JOIN {src_table} b ON a.package_id = b.id "
						)
					condition = Condition()
					if vul_id:
						vul_year, vul_num = SplitVulId(vul_id)
						condition.Add(f"a.vul_year = {vul_year} AND "
							f"a.vul_num = '{vul_num}'")
					if package:
						condition.Add(f"b.name = '{package}'")
					condition.Add(Condition('', type_of_issues,
						self.context.fix, years=self.context.years))
					query += condition.Where() + ' ' \
						'GROUP BY a.vul_year, a.vul_num, b.name'
					issues = self.db.Cursor(query).fetchall()
					if not issues:
						continue
					for vul_year, vul_num, package, fix in issues:
						entries[(branch, package, vul_year, vul_num)]= {
							'desc': self.__QueryVulDesc(type_of_issues, vul_year, vul_num),
							'fix' : FilterFix(fix)
							}

		return entries

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get and print a map of package names (pack_name >> cpe)

	def PackagesMap(self, type_of_issues, branch):

		if not self.context:
			raise FatalError('Context is not defined')
			return []

		table = branch + '_src'

		if not self.db.TableExists(table):
			raise FatalError(f'Table "{table}" doesn\'t exist')
			return []

		mapped_name = f"{type_of_issues}_name"
		cols = ["name", mapped_name]
		cond = f"{mapped_name} <> '' "

		if self.context.packages:
			cond += "AND name in (" + \
				', '.join(f"'{p}'" for p in self.__Packages(branch)) + \
				")"

		cursor = self.db.Select(cols, table, cond)

		if not cursor:
			return tuple()

		return cursor.fetchall()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print summary of a map of package names

	def MapCount(self, branch, type_of_issues):

		table = branch + '_src'
		n_packages = self.db.CountRows(table)
		n_products = self.db.CountRows(table,
			cond=f"{type_of_issues}_name <> ''")

		return (n_packages, n_products)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Compare given branches

	def CompareBranches(self, type_of_issues, branches, printing=True):

		diff = defaultdict(list)
		if not self.context:
			raise FatalError('Context is not defined')
			return diff

		# At lest two available branches to compare
		if len(branches) < 2:
			raise FatalError('Need at least two branch names')
			return diff

		# Selecting the volumes to handle
		years = ReviseCVEVolumes(self.context.years, self.printer)[0:-1] \
			if self.context.years else []

		# This function can run in silent mode
		silent0 = self.printer.IsSilent()
		self.printer.Silent(not printing)

		# We need opposite subtypes of issues for first branch of the list
		#	and the rest
		(fix_status, opposite_fix_status) = ('fixed', 'unfixed') \
			if not self.context.fix or self.context.fix in ('fixed', 'all') \
			else ('unfixed', 'fixed')

		# Forming name of the issues table - nested func
		def IssuesTable(branch, type_of_issues):
			return f'{branch}_{type_of_issues}_issues'
		# - nested func

		# Getting the issues data for all the mentioned branches
		self.printer.LineBegin('Searching for vulnerabilities that '
			f'{fix_status} in {", ".join(branches[1:])} and isn\'t '
			f'{fix_status} in {branches[0]}')
		issues_table1 = IssuesTable(branches[0], type_of_issues)
		src_table1 = branches[0] + '_src'
		for branch in branches[1:]:
			issues_table2 = IssuesTable(branch, type_of_issues)
			src_table2 = branch + '_src'
			condition = Condition('', type_of_issues, fix_status,
				None, years, issues_table1)
			condition.Add(Condition(packages=self.__Packages(branch),
				target=src_table1))
			condition.Add(Condition('', type_of_issues, opposite_fix_status,
				None, years, issues_table2))
			condition.Add(Condition(packages=self.__Packages(branch),
				target=src_table2))
			# Forming a query to select issues of a branch from branches[1:] and
			# replacing a name of ver col of a src table of that branch in that
			# query with a name of ver col of a src table of a branches[0] branch
			query_issues_of_branch = QueryToFetchIssues(branch, type_of_issues,
				acl=True, fix=True, scores=True, disputed=True, concat_year_num=True
				).replace(f'{src_table2}.ver', f'{src_table1}.ver')
			query = (f"{query_issues_of_branch} "
				f"INNER JOIN {issues_table1} "
				f"ON  {issues_table1}.vul_year = {issues_table2}.vul_year "
				f"AND {issues_table1}.vul_num  = {issues_table2}.vul_num "
				f"AND {issues_table1}.vul_ver  = {issues_table2}.vul_ver "
				f"INNER JOIN {src_table1} "
				f"ON  {issues_table1}.package_id = {src_table1}.id "
				f"AND {src_table1}.name = {src_table2}.name "
				f"AND {src_table1}.{type_of_issues}_name = {src_table2}.{type_of_issues}_name "
				f"{condition.Where()}")
			query += self.__CustomOrder(type_of_issues=type_of_issues,
				branch=branch)
			ret = self.db.Cursor(query)
			if ret == None:
				self.printer.Silent(silent0)
				raise FatalError('DB request failed')
				break
			diff[branch] = ret.fetchall()

		n_issues = sum([len(issues) for issues in diff.values()])
		if n_issues > 0:
			self.printer.LineEnd(f'[{n_issues} FOUND]')
		else:
			self.printer.LineEnd('[NONE]')

		self.printer.Silent(silent0)

		return diff

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Keeping a track of unfixed issues and a map of package names

	# Get a date-time list out of a given history table name
	def __GetDateTimeOfHistoryTable(self, history_table_name):
		return [int(el) for el in history_table_name.split('_')[-6:]]

	# Get a name of a most recent history table
	def __GetRecentHistoryTableName(self, type_of_history, type_of_issues,
			branch):

		recent_history_table_name = ''
		recent_history_table_date = [0] * 6
		history_tables = \
			self.db.ShowHistoryTables(type_of_history, type_of_issues, branch)

		for table_name in history_tables:
			table_date_time = self.__GetDateTimeOfHistoryTable(table_name)
			if table_date_time > recent_history_table_date:
				recent_history_table_date = table_date_time
				recent_history_table_name = table_name

		return recent_history_table_name

	# Get unfixed issues either from issues tables or from history tables
	def __QueryUnfixedIssues(self, type_of_issues, branch,
			select_from_history_tables=False):

		cursor = None

		if select_from_history_tables:
			history_table = \
				self.__GetRecentHistoryTableName('issues', type_of_issues, branch)
			if history_table:
				condition = Condition(type_of_issues=type_of_issues,
					packages=self.__Packages(branch), years=self.context.years)
				vul_marker = VUL_MARKERS.get(type_of_issues, '')
				cols = [
					f"CONCAT('{vul_marker}', vul_year, '-', vul_num) vul_id",
					"name", f"{type_of_issues}_name", "ver", "rel", "vul_ver"
					]
				cursor = self.db.Select(cols, history_table, condition.What())
		else:
			src_table = f'{branch}_src'
			issues_table = f'{branch}_{type_of_issues}_issues'
			# Forming a condition for packages
			condition = Condition(type_of_issues=type_of_issues,
				packages=self.__Packages(branch), target=src_table)
			# Adding a condition for years
			condition.Add(Condition(type_of_issues=type_of_issues,
				years=self.context.years, target=issues_table))
			order = self.__CustomOrder(type_of_issues=type_of_issues,
				branch=branch)
			condition.Add(f"fix = 'V' {order}")
			cursor = self.db.QueryIssues(branch, type_of_issues,
				condition.What(), fix=True, disputed=True, concat_year_num=True)

		return cursor.fetchall() if cursor else []

	# Get a diff between issues-history tables and issues tables
	def __QueryNewUnfixedIssues(self, type_of_issues):

		# Convert a list of entries to a dict of entries
		#	with a (vul_id, name, <type_of_issues>_name) tuple as the key
		def ListStateToDictState(list_state):
			dict_state = dict()
			for entry in list_state:
				dict_state[tuple(entry[0:3])] = tuple(entry[3:])
			return dict_state
		# - nested func

		new_unfixed_issues = defaultdict(list)
		for branch in self.context.branches:
			current_unfixed_issues = ListStateToDictState(
					self.__QueryUnfixedIssues(type_of_issues, branch, False))
			recent_unfixed_issues  = ListStateToDictState(
					self.__QueryUnfixedIssues(type_of_issues, branch, True))
			for k, v in current_unfixed_issues.items():
				if k not in recent_unfixed_issues:
					new_unfixed_issues[branch].append(k + v)

		return new_unfixed_issues

	# Get a diff between map-history tables and src tables
	def __QueryPackageNamesMapChanges(self, type_of_product_names):

		changes = defaultdict(list)

		for branch in self.context.branches:
			src_table = branch + '_src'
			history_table = self.__GetRecentHistoryTableName('map', '', branch)
			if not history_table:
				continue
			query = ("SELECT a.name, "
				"b.{s}_name, a.{s}_name ".format(s=type_of_product_names) +
				f"FROM {src_table} a INNER JOIN {history_table} b ON ("
					"a.name = b.name AND (" +
						"a.{s}_name <> b.{s}_name OR "
						"(a.{s}_name IS NOT NULL AND b.{s}_name IS NULL) OR "
						"(b.{s}_name IS NOT NULL AND a.{s}_name IS NULL)"
						")".format(s=type_of_product_names) +
					") "
				"ORDER BY a.name")
			# Sorting newly found, newly lost and newly changed name matches
			fnd, lst, chg = [], [], []
			for row in self.db.Cursor(query):
				if not row[1]:
					fnd.append((row[0], '[NONE]', row[2]))
				elif not row[2]:
					lst.append((row[0], row[1], '[NONE]'))
				else:
					chg.append(row)
			for rows in (fnd, lst, chg):
				for row in rows:
					changes[branch].append([row[0], f'{row[1]} >> {row[2]}'])

		return changes

	# Get new unfixed issues or changes in maps of package names
	def History(self, target, specification):

		if target == 'issues':
			return self.__QueryNewUnfixedIssues(specification)

		return self.__QueryPackageNamesMapChanges(specification)
