#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from copy               import deepcopy
from ax.ver             import MergeVersions
from cve_manager        import defines as df
from cve_manager.common import UntaggedRefs
from cve_monitor.common import PackagesList

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Convert given software list to a more human-readable format

def ReadableSoftwareList(software_list):

	res = ''

	for conjunction in software_list.split():
		if conjunction.count('&&') > 0:
			conjunction = '{' + conjunction.replace('&&', ' && ') + '}'
		res += conjunction + ' '

	return res[:-1]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class Renderer():

	def __init__(self, printer):

		self.printer = deepcopy(printer)
		self.context = None
		self.RESTRICTED_COLS = {2: 30, 3: 25}
		self.RESTRICTED_COLS_SORTED = sorted(
			[(k, v) for k, v in self.RESTRICTED_COLS.items()],
			key=lambda t: t[0]
			)
		self.EXCLUDABLE_COLS = {'ACL'}
		self.FIX_STATUS = {
			'e': 'excluded',
			'c': 'changelog',
			'b': 'bigger ver',
			'L': 'lesser ver',
			'V': '[NO FIX]',
			'x': 'false complex vulnerability',
			'?': 'unresolved',
			}

	def SetContext(self, context):

		if any(not hasattr(context, attr) for attr
				in ('fix', 'packages', 'years', 'patch', 'group', 'distro')):
			return False

		self.context = context

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a list of packages of a specified branch

	def __Packages(self, branch=None):
		
		return PackagesList(self.context.packages, branch)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Calculate columns width needed to fit every piece of issues table

	def __ColumnsWidth(self, list_of_lists, ignore_last_column=False):

		max_width = []
		ncols = min(map(len, list_of_lists))

		for i in range(0, ncols):
			max_width.append(0)

		# Width of last column may not be required
		if ignore_last_column:
			ncols -= 1
		for l in list_of_lists:
			for i in range(0, ncols):
				width = len(l[i])
				if width > max_width[i]:
					if i not in self.RESTRICTED_COLS or \
							width < self.RESTRICTED_COLS[i]:
						max_width[i] = width

		return max_width

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print names of the branches where particular vulnerability is fixed

	def __Fixes(self, fixes):

		# Decode given status string
		def DecodeStatus(status):
			return ', '.join([self.FIX_STATUS[c] for c in status])
		# - nested func

		if fixes:
			if type(fixes) is dict:
				for branch, fix in fixes.items():
					self.printer.LineEnd(f'Fix in {branch}: {DecodeStatus(fix)}')
			elif type(fixes) is str:
				self.printer.LineEnd(f'Fix: {DecodeStatus(fixes)}')

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Converting score from float to string

	def __ScoreStr(self, score):
		return str(score) if score > 0.01 else '-'

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Converting string representation of a range of vulnerable versions

	def __ConvVulVer(self, vul_versions):

		vul_ver_conv = ''

		vul_versions = vul_versions.split()
		if self.context.group:
			vul_versions = MergeVersions(vul_versions)

		for vul_ver in vul_versions:
			vul_ver = vul_ver.strip()
			if not vul_ver:
				continue
			if vul_ver[0] not in ('[', '(') or vul_ver[-1] not in (']', ')'):
				if vul_ver == '-':
					vul_ver = '*'
				vul_ver_conv += vul_ver + ', '
				continue
			vul_ver_range = vul_ver.strip('[()]').split(';')
			range_left  = vul_ver_range[0]
			range_right = vul_ver_range[1]
			start_incl = vul_ver[0]  == '['
			end_incl   = vul_ver[-1] == ']'
			if not range_left and not range_right:
				vul_ver_conv += '*, '
				continue
			if range_left:
				if not range_right:
					vul_ver_conv += '>'
					if start_incl:
						vul_ver_conv += '='
					vul_ver_conv += range_left
				else:
					vul_ver_conv += vul_ver[0] + range_left + '; '
			if range_right:
				if not range_left:
					vul_ver_conv += '<'
					if end_incl:
						vul_ver_conv += '='
					vul_ver_conv += range_right
				else:
					vul_ver_conv += range_right + vul_ver[-1]
			vul_ver_conv += ', '

		return vul_ver_conv[:-2] if vul_ver_conv else ''

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Form a full name for a monitored target (branch or distro)

	def __TargetStr(self, branch):

		return f'{branch}:{self.context.distro} distro' if self.context.distro \
			else f'{branch} branch'

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print title of a report

	def Title(self, type_of_issues):

		self.printer.Div()

		title = '\n.:: CVE-MANAGER ANALYSIS OF {t} DATA ::.\n' \
			.format(t=type_of_issues.upper())
		self.printer.LineEnd(title)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print given entries (with the given ids) of the nvd_vul_import table

	def NVDEntries(self, full_desc, fixes, branch, package):

		self.printer.Div()

		for (cve_id, software_list, published, modified, cvss_score_v2,
				cvss_score_v3, cwe_id, refs, patch, summary) in full_desc:
			if branch:
				self.printer.LineEnd(f'Branch: {branch}')
			if package:
				self.printer.LineEnd(f'Package: {package}')
			self.printer.Multiline([
				f'CVE id: {cve_id}',
				f'Software list: {ReadableSoftwareList(software_list)}',
				f'Published: {published}',
				f'Modified: {modified}',
				f'CWE id: {cwe_id}',
				f'CVSS v2 score: {self.__ScoreStr(cvss_score_v2)}',
				f'CVSS v3 score: {self.__ScoreStr(cvss_score_v3)}',
				f'Patch refs: {patch.replace(" ", ", ")}',
				f'Non patch refs: {UntaggedRefs(refs, ", ")}',
				f'Summary: {summary}'
				])
			self.__Fixes(fixes)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print given entries (with the given ids) of the fstec_vul_import table

	def FSTECEntries(self, full_desc, fixes, branch, package):

		self.printer.Div()

		for (vul_id, vul_name, description, software_list, environment, cwe_id,
				published, cvss_score_v2, cvss_score_v3, severity, solution,
				vul_status, exploit_status, fix_status, sources, other_ids) \
				in full_desc:
			if branch:
				self.printer.LineEnd(f'Branch: {branch}')
			if package:
				self.printer.LineEnd(f'Package: {package}')
			self.printer.Multiline([
				f'Vul id: {vul_id}',
				f'Other ids: {other_ids}',
				f'Description: {description}',
				f'Software list: {ReadableSoftwareList(software_list)}',
				f'Environment: {environment}',
				f'Published: {published}',
				f'CWE id: {cwe_id}',
				f'CVSS v2 score: {self.__ScoreStr(cvss_score_v2)}',
				f'CVSS v3 score: {self.__ScoreStr(cvss_score_v3)}',
				f'Severity: {severity}',
				f'Solution: {solution}',
				f'Vul status: {vul_status}',
				f'Exploit status: {exploit_status}',
				f'Fix status: {fix_status}',
				f'Sources: {sources}'
				])
			self.__Fixes(fixes)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print a header for an issues report

	def __IssuesHeader(self, n_issues, issues_type, branch, this_is_new_issues):

		self.printer.Div()

		msg = branch + \
			(f':{self.context.distro} ' if self.context.distro else ' ') + \
			('new ' if this_is_new_issues else '') + \
			f'{self.context.fix} {issues_type.upper()} issues'

		if self.context.packages and not self.context.distro:
			packages = self.__Packages(branch)
			msg += f' for {", ".join(packages)} ' \
				f'package{"s" if len(packages) > 1 else ""}'

		msg += f', {n_issues} ' + ('vulnerable packages' if self.context.group
			else 'vulnerabilities') + ' total'

		if self.context.years:
			msg += f' (year{"s" if len(self.context.years) > 1 else ""} ' \
				f'{", ".join(self.context.years)})'

		self.printer.LineEnd(f'\n{msg}:')

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print particular issues of particular type and branch

	def Issues(self, issues_type, branch, issues, suggestions=None,
			this_is_new_issues=False, display_header=True):

		if not self.context:
			self.printer.Err('Context is not defined')
			return

		# Printing the header
		if display_header:
			n_issues = len(issues) if issues else 0
			self.__IssuesHeader(n_issues, issues_type, branch,
				this_is_new_issues)

		self.printer.LineEnd('')

		if not issues:
			self.printer.LineEnd('-- None --')
			return

		# List of branches that could be suggested as the source
		#       of replacements for vulnerable packages
		branches = [b for b in self.context.all_branches if b != branch]

		# Exclude columns that has less than 5% of non-empty values
		def ExcludeEmptyColumns(cols, order_of_cols, grouped_issues):
			counter = [0] * len(grouped_issues[0])
			for entry in grouped_issues[1:]:
				for i, el in enumerate(entry):
					if el:
						counter[i] += 1
			min_percentage = (len(grouped_issues) - 1) * 0.05;
			for i, el in enumerate(counter):
				if el < min_percentage and i in order_of_cols and \
						cols.get(i, '') in self.EXCLUDABLE_COLS:
					order_of_cols.remove(i)
			return order_of_cols
		# - nested func

		# Calculate a max allowed package name length
		def MaxPackageNameLen(issues):
			max_len = 0
			for entry in issues:
				name, ver = entry[1], entry[3]
				l = len(name) + 1 + len(ver) + 1 + 5
				if l > max_len:
					max_len = l
			return max_len
		# - nested func

		# Trim a release-part of a package name if this name is too long
		def TrimRel(name, ver, rel):
			package_name = f'{name}-{ver}-{rel}'
			if len(package_name) > max_package_name_len and len(rel) > 5:
				return '-'.join([name, ver, f'{rel[0]}..{rel[-2:]}'])
			return package_name
		# - nested func

		# Form a string of cure suggestions
		def Cure(name, vul_id):
			if not suggestions:
				return []
			cure = []
			if self.context.group:
				for b in branches:
					fixed_issues = suggestions.get((name, b), 0)
					if fixed_issues <= 0:
						continue
					cure.append(b + (f'({fixed_issues} fixed)'
						if fixed_issues < int(vul_id) else ''))
			else:
				for b in branches:
					if suggestions.get((vul_id, b), 1) == 0:
						cure.append(b)
			return ', '.join(cure)
		# - nested func

		there_are_disputed_vulnerabilities = False

		# Group fields of the issues entry - nested func
		def GroupIssuesEntry(issues_entry):
			nonlocal there_are_disputed_vulnerabilities
			nvul_or_vulid, name, map_name, ver, rel, vul_ver, score2, score3, disputed, fix, acl = \
				issues_entry[:-1] if self.context.patch else issues_entry
			nvul_or_vulid_str = str(nvul_or_vulid)
			if disputed:
				there_are_disputed_vulnerabilities = True
				nvul_or_vulid_str = '{*}' + nvul_or_vulid_str
			patch  = issues_entry[-1].replace(' ', ', ') if self.context.patch else ''
			scores = self.__ScoreStr(score2) + ' / ' + self.__ScoreStr(score3)
			issues_entry = [
				nvul_or_vulid_str,             # 0 - QTY or CVE ID
				TrimRel(name, ver, rel),       # 1 - PACKAGE NAME-VER-REL
				map_name,                      # 2 - MAPPED NAME
				self.__ConvVulVer(vul_ver),    # 3 - VULNERABLE VERSIONS
				scores,                        # 4 - CVSS2/3 or MAX CVSS2/3
				fix,                           # 5 - FIX
				acl,                           # 6 - ACL
				Cure(name, nvul_or_vulid),     # 7 - CURE
				patch,                         # 8 - PATCH
				]
			return issues_entry
		# - nested func

		# Defining an order of columns
		cols = {
			0: 'QTY' if self.context.group else 'CVE ID',
			1: 'PACKAGE NAME-VER-REL',
			2: 'MAPPED NAME',
			3: 'VULNERABLE VER',
			4: 'MAX CVSS2/3' if self.context.group else 'CVSS2/3',
			5: 'FIX',
			6: 'ACL',
			7: 'CURE',
			8: 'PATCH',
			}
		order_of_cols  = [1, 2, 3, 0] if self.context.group else [0, 1, 2, 3]
		order_of_cols += [4]
		order_of_cols += [5] if self.context.fix != 'unfixed' else []
		order_of_cols += [6]
		order_of_cols += [7] if suggestions else []
		order_of_cols += [8] if self.context.patch else []

		# Processing the entries of the report table
		footnotes = {}
		if not issues:
			self.printer.LineEnd('-- None --')
		else:
			# Defining a max allowed package name length
			max_package_name_len = MaxPackageNameLen(issues)
			# Forming a body of a report table
			grouped_issues = [[cols.get(i, '') for i in range(len(cols))]]
			for issues_entry in issues:
				grouped_issues.append(GroupIssuesEntry(issues_entry))
			# Excluding empty and almost empty columns
			order_of_cols = ExcludeEmptyColumns(cols, order_of_cols, grouped_issues)
			# Calculating required width of the columns
			cols_width = self.__ColumnsWidth(grouped_issues, True)
			# Printing the entries
			for issues_entry in grouped_issues:
				# Too long lists of vulnerable versions will be placed in the footnotes
				for restricted_col, max_len in self.RESTRICTED_COLS_SORTED:
					item = issues_entry[restricted_col]
					if len(item) > (max_len - 1):
						if item not in footnotes:
							footnotes[item] = len(footnotes) + 1
						issues_entry[restricted_col] = f'{{{footnotes[item]}}}'
				entry_str = ''
				for i in order_of_cols[:-1]:
					if i < 0:
						continue
					entry_str += issues_entry[i].ljust(cols_width[i]) + ' | '
				self.printer.LineEnd(entry_str + issues_entry[order_of_cols[-1]])

		# Printing the footnotes
		if footnotes or there_are_disputed_vulnerabilities:
			self.printer.LineEnd('')
			if there_are_disputed_vulnerabilities:
				desc = 'some vulnerabilities of this package are disputed' \
					if self.context.group \
					else 'this vulnerability is disputed'
				self.printer.LineEnd(f'{{*}} {desc}')
		for item, index in sorted(footnotes.items(), key=lambda kv: kv[1]):
			self.printer.LineEnd(f'{{{index}}} {item}')

		self.printer.LineEnd('')

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Printing the diff between branches

	def Diff(self, issues_type, branch_x, branch_y, diff):

		if not self.context:
			self.printer.Err('Context is not defined')
			return

		self.printer.Div()
		self.printer.LineBreak()
		msg = (
			f'{self.context.fix.capitalize()} vulnerabilities '
			f'in {self.__TargetStr(branch_x)} that are not '
			f'{self.context.fix if self.context.fix != "all" else "found"} '
			f'in {self.__TargetStr(branch_y)}, {len(diff)} total:'
			)
		self.printer.LineEnd(msg)

		self.Issues(issues_type, branch_y, diff, display_header=False)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print a given map of package names of a given branch

	def Map(self, issues_type, branch, pairs):

		if not self.context:
			self.printer.Err('Context is not defined')
			return

		self.printer.Div()

		msg = branch
		if self.context.packages:
			msg += f'map of {{{", ".join(self.__Packages(branch))}}} packages'
		else:
			msg += f'package names to {issues_type.upper()} map'

		self.printer.LineEnd(f'\n{msg}:\n')

		if pairs:
			for package_name, product_name in pairs:
				self.printer.LineEnd(f'{package_name} >> {product_name}')
		else:
			self.printer.LineEnd('-- None --')

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print a given list of changes of a map of package names of a given branch

	def ChangesOfMap(self, type_of_product_names, branch, changes):

		# Print given entries
		def PrintEntries(entries):
			for entry in entries:
				entry_str = ''
				for index, content in enumerate(entry[:-1]):
					entry_str += content.ljust(cols_width[index]) + ' | '
				self.printer.LineEnd(entry_str + entry[-1])
		# - nested func

		self.printer.Div()

		msg = f'Changes of a map of package names of {branch} branch'
		self.printer.LineEnd(f'\n{msg}:\n')

		if not changes:
			self.printer.LineEnd('-- None --')
			return

		header = [['PACKAGE NAME', type_of_product_names.upper() + ' NAME']]

		# Calculating required width of the columns
		cols_width = self.__ColumnsWidth(changes + header, True)

		# Rendering the header and then the body of a table
		PrintEntries(header)
		PrintEntries(changes)
		self.printer.LineBreak()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Printing a share of mapped packages out of total number of packages

	def MapCount(self, n_packages, n_products):

		msg = f'There {"is" if n_products == 1 else "are"} ' \
			f'{n_products} mapped package{"" if n_products == 1 else "s"} ' \
			f'out of {n_packages}'
		self.printer.LineEnd(f'\n{msg}\n')
